/* -*- C -*-
 * Copyright (c) 1999, 2000 Allan Saddi <asaddi@philosophysw.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: config.c,v 1.21 2000/08/07 02:57:34 asaddi Exp $
 */

#include <sys/types.h>

#include <ctype.h>
#include <limits.h>
#include <pwd.h>
#include <regex.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include <assert.h>

#include "util.h"
#include "list.h"
#include "newsfish.h"
#include "config.h"

#ifndef lint
static const char rcsid[] =
	"$Id: config.c,v 1.21 2000/08/07 02:57:34 asaddi Exp $";
#endif /* !lint */

#define LINEBUF_LEN 1024

struct List ServerList, NewsgroupList, WildNewsgroupList;

static const char *
basename (const char *name)
{
  int i;

  assert (name != NULL);

  for (i = strlen (name) - 1; i >= 0 && name[i] != '/'; i--);
  i++;
  return &name[i];
}

/* Stuff that should probably be in config.c */

/* Add a server to our list. */
struct Server *
AddServer (const char *server)
{
  struct Server *servInfo;

  servInfo = mymalloc (sizeof (*servInfo));

  memset (servInfo, 0, sizeof (*servInfo));
  strlcpy (servInfo->name, server, sizeof (servInfo->name));
  servInfo->port = DefaultPort;

  AddTail (&ServerList, &servInfo->node);

  return servInfo;
}

/* Find a particular server. */
struct Server *
FindServer (const char *server)
{
  struct Server *servInfo;

  FOR_EACH_NODE_IN_LIST (&ServerList, servInfo)
    if (!strcasecmp (servInfo->name, server))
      return servInfo;

  return NULL;
}

/* Add a newsgroup to our list. */
struct Newsgroup *
AddNewsgroup (const char *newsgroup)
{
  struct Newsgroup *ngInfo;

  ngInfo = mymalloc (sizeof (*ngInfo));

  strlcpy (ngInfo->name, newsgroup, sizeof (ngInfo->name));
  ngInfo->lastArticle = 0;
  InitList (&ngInfo->actionList);
  ngInfo->wildType = WILD_TYPE_NONE;

  AddTail (&NewsgroupList, &ngInfo->node);

  return ngInfo;
}

/* Find a particular newsgroup in our list. */
struct Newsgroup *
FindNewsgroup (const char *newsgroup)
{
  struct Newsgroup *ngInfo;

  FOR_EACH_NODE_IN_LIST (&NewsgroupList, ngInfo)
    if (!strcmp (ngInfo->name, newsgroup))
      return ngInfo;

  return NULL;
}

/* Returns the type of a wildcard newsgroup. */
static enum WildcardType
getWildNewsgroupType (const char *newsgroup)
{
  int patLen;
  int wildCount;
  const char *line;

  assert (newsgroup != NULL);
  assert (newsgroup[0] != '\0');

  patLen = strlen (newsgroup);

  /* Handle the easy case first. */
  if (!strcmp (newsgroup, "*"))
    return WILD_TYPE_ALL;

  /* Count the number of wildcards. */
  wildCount = 0;
  line = newsgroup;
  while (*line) {
    if (*line == '*')
      wildCount++;
    line++;
  }

  /* This better be a wildcard newsgroup! */
  assert (wildCount > 0);

  /* First and last character a wildcard? */
  if (wildCount == 2 && (newsgroup[0] == '*' && line[-1] == '*') && patLen > 2)
    return WILD_TYPE_MIDDLE;
  /* First or last character a wildcard? */
  else if (wildCount == 1) {
    if (newsgroup[0] == '*')
      return WILD_TYPE_SUFFIX;
    else if (line[-1] == '*')
      return WILD_TYPE_PREFIX;
  }

  /* Something we don't support. */
  return WILD_TYPE_NONE;
}

/* Add a wildcard newsgroup to our list. Right now, we only handle four types
 * of wildcard newsgroups. (Maybe until we find a suitable patter matcher --
 * fnmatch()?) These are:
 *   1. *
 *   2. *.newsgroup.suffix
 *   3. newsgroup.prefix.*
 *   4. *.newsgroup.middle.*
 */
struct Newsgroup *
AddWildNewsgroup (const char *newsgroup)
{
  struct Newsgroup *ngInfo;
  enum WildcardType wildType;
  int i;

  assert (newsgroup != NULL);

  ngInfo = mymalloc (sizeof (*ngInfo));

  /* See if this is a valid wildcard newsgroup. */
  if ((wildType = getWildNewsgroupType (newsgroup)) == WILD_TYPE_NONE)
    return NULL;

  /* Strip out all wildcard chars. We can do this since we know
     the newsgroup pattern is already valid. */
  i = 0;
  while (*newsgroup && i < sizeof (ngInfo->name) - 1) {
    if (*newsgroup != '*')
      ngInfo->name[i++] = *newsgroup;
    newsgroup++;
  }
  ngInfo->name[i] = '\0';

  ngInfo->lastArticle = 0;
  InitList (&ngInfo->actionList);
  ngInfo->wildType = wildType;

  AddTail (&WildNewsgroupList, &ngInfo->node);

  return ngInfo;
}

/* Write out the newsRC file. */
int
WriteNewsRC (const char *newsrc)
{
  int success = 0;
  FILE *rcFile;
  struct Newsgroup *ngInfo;

  if ((rcFile = myfopen (newsrc, "w"))) {
    success = 1;

    FOR_EACH_NODE_IN_LIST (&NewsgroupList, ngInfo) {
      /* Minor adjustment, since we're marking what we've already *read*. */
      if (ngInfo->lastArticle > 0)
	ngInfo->lastArticle--;
      /* So the output doesn't look weird. */
      if (ngInfo->lastArticle < 2)
	ngInfo->lastArticle = 1;
      fprintf (rcFile, "%s: 1-%d\n", ngInfo->name, ngInfo->lastArticle);
    }

    fclose (rcFile);
  }

  return success;
}

static void
parseNewsRCLine (char *line)
{
  char newsgroup[LINEBUF_LEN];
  int articleLow, articleHigh;
  int i;
  struct Newsgroup *ngInfo;

  if (sscanf (line, "%[^:]:%d -%d", newsgroup, &articleLow, &articleHigh)
      == 3) {
    for (i = 0; newsgroup[i] && !isspace ((int) newsgroup[i]); i++);
    newsgroup[i] = '\0';

    if ((ngInfo = FindNewsgroup (newsgroup))) {
      /* Remember to adjust the article number... we want to know what
	 to read *next*. */
      ngInfo->lastArticle = articleHigh + 1;
    }
    else {
      /* Just create a blank newsgroup to hold this last article. */
      ngInfo = AddNewsgroup (newsgroup);
      ngInfo->lastArticle = articleHigh + 1;
    }
  }
}

int
ReadNewsRC (const char *newsrc)
{
  int success = 0;
  FILE *rcFile;
  char lineBuf[LINEBUF_LEN], *line;
  int i;

  if ((rcFile = myfopen (newsrc, "r"))) {
    success = 1;

    while (fgets (lineBuf, sizeof (lineBuf), rcFile)) {
      line = lineBuf;

      while (*line && isspace ((int) *line))
	line++;

      if (!(*line) || *line == '#')
	continue;

      i = strlen (line);
      if (i > 0 && line[i - 1] == '\n')
	line[i - 1] = '\0';

      parseNewsRCLine (line);
    }

    fclose (rcFile);
  }

  return success;
}

int
ReadConfig (const char *conf)
{
  int success = 0;
  FILE *rcFile;
  char lineBuf[LINEBUF_LEN], *line;
  int lineNo = 0;
  enum {
    NEWSGROUP,
    CONDITION,
  } state = NEWSGROUP;

  int i;

  struct Newsgroup *ngInfo;
  struct Action *actInfo;
  struct Condition *condInfo;
  int reErrCode;

  int rereadLine;

  if ((rcFile = myfopen (conf, "r"))) {
    success = 1;
    ngInfo = NULL;
    actInfo = NULL;

    while (fgets (lineBuf, sizeof (lineBuf), rcFile)) {
      lineNo++;

      do {
	rereadLine = 0;
	line = lineBuf;

	while (*line && isspace ((int) *line))
	  line++;

	if (*line == '#')
	  continue;

	/* Strip any trailing whitespace. */
	i = strlen (line) - 1;
	for (; i > 0 && isspace ((int) line[i]); i--);
	line[++i] = '\0';

	if (!(*line))
	  continue;

	switch (state) {
	case NEWSGROUP:
	  for (i = 0;
	       line[i] && !isspace ((int) line[i]) && line[i] != ':';
	       i++);
	  if (line[i] == ':') {
	    line[i] = '\0';
	    i++;
	    for (; line[i] && isspace ((int) line[i]); i++);
	    if (line[i]) {
	      fprintf (stderr, "%s: %s:%d: bad newsgroup line\n",
		       PROGRAM_NAME, basename (conf), lineNo);
	      continue;
	    }

	    if (!(*line)) {
	      fprintf (stderr, "%s: %s:%d: bad newsgroup line\n",
		       PROGRAM_NAME, basename (conf), lineNo);
	      continue;
	    }

	    /* Is this a wildcard newsgroup? */
	    if (strchr (line, '*')) {
	      /* Add this wildcard newsgroup, even if it's a duplicate. */
	      if ((ngInfo = AddWildNewsgroup (line))) {
		/* Add the first action to this wildcard newsgroup. */
		actInfo = mymalloc (sizeof (*actInfo));

		/* Initialize this action. */
		InitList (&actInfo->conditionList);
		actInfo->type = ACTION_TYPE_NONE;
		actInfo->file[0] = '\0';
		actInfo->sizeLimit = GlobalSizeLimit;

		/* Link it into the current newsgroup. */
		AddTail (&ngInfo->actionList, &actInfo->node);
	      }
	      else {
		/* Not a valid wildcard newsgroup. */
		fprintf (stderr, "%s: %s:%d: bad wildcard newsgroup\n",
			 PROGRAM_NAME, basename (conf), lineNo);
		continue;
	      }
	    }
	    else {
	      /* Prepare to add conditions to this newsgroup. */
	      if (!(ngInfo = FindNewsgroup (line))) {
		/* Newsgroup doesn't exist. Add this new newsgroup. */
		ngInfo = AddNewsgroup (line);
	      }
	      assert (ngInfo != NULL);

	      /* Add a new action to this newsgroup. (It may be the first) */
	      actInfo = mymalloc (sizeof (*actInfo));

	      /* Initialize this action. */
	      InitList (&actInfo->conditionList);
	      actInfo->type = ACTION_TYPE_NONE;
	      actInfo->file[0] = '\0';
	      actInfo->sizeLimit = GlobalSizeLimit;

	      /* Link it into the current newsgroup. */
	      AddTail (&ngInfo->actionList, &actInfo->node);
	    }

	    state = CONDITION;
	  }
	  else
	    fprintf (stderr, "%s: %s:%d: bad newsgroup line\n",
		     PROGRAM_NAME, basename (conf), lineNo);
	  break;

	case CONDITION:
	  if (*line == '*' && line[1] == ' ') {
	    /* Condition. */
	    line += 2;
	    while (*line && isspace ((int) *line))
	      line++;
	    if (*line) {
	      /* Add this condition. */
	      assert (ngInfo != NULL);
	      assert (actInfo != NULL);

	      condInfo = mymalloc (sizeof (*condInfo));

	      /* Set the invert flag appropriately. */
	      if ((condInfo->invert = (*line == '!'))) {
		/* Skip the ! and any white spaces. */
		line++;
		while (*line && isspace ((int) *line))
		  line++;
	      }

	      if (*line) {
		if (!(reErrCode = regcomp (&condInfo->regexp, line,
					   REG_EXTENDED | REG_ICASE | REG_NOSUB |
					   REG_NEWLINE)))
		  AddTail (&actInfo->conditionList, &condInfo->node);
		else {
		  regerror (reErrCode, &condInfo->regexp, lineBuf,
			    sizeof (lineBuf));
		  fprintf (stderr, "%s: %s:%d: bad regexp -- %s\n",
			   PROGRAM_NAME, basename (conf), lineNo, lineBuf);
		  free (condInfo);
		}
	      }
	      else
		fprintf (stderr, "%s: %s:%d: bad condition line\n",
			 PROGRAM_NAME, basename (conf), lineNo);
	    }
	    else
	      fprintf (stderr, "%s: %s:%d: bad condition line\n",
		       PROGRAM_NAME, basename (conf), lineNo);
	  }
	  else {
	    /* Destination file. */
	    assert (ngInfo != NULL);
	    assert (actInfo != NULL);

	    /* Are we actually handling a wildcard newsgroup? */
	    if (ngInfo->wildType != WILD_TYPE_NONE) {
	      /* If so, we're done. */
	      ngInfo = NULL;
	      actInfo = NULL;

	      state = NEWSGROUP;
	      rereadLine = 1;
	      continue;
	    }

	    /* Figure out what kind of action to take. */
	    if (*line == '|') {
	      /* Pipe it to a program. */
	      line++;
	      while (*line && isspace ((int) *line))
		line++;
	      if (*line) {
		strlcpy (actInfo->file, line, sizeof (actInfo->file));
		actInfo->type = ACTION_TYPE_PIPE;
	      }
	      else {
		/* Sorry, we don't accept just a pipe symbol by itself. */
		fprintf (stderr, "%s: %s:%d: bad pipe action\n",
			 PROGRAM_NAME, basename (conf), lineNo);
	      }
	    }
	    else {
	      /* A regular file. Simply copy it verbatim. */
	      strlcpy (actInfo->file, line, sizeof (actInfo->file));
	      actInfo->type = ACTION_TYPE_FILE;
	    }

	    /* So we know if we abort early. */
	    ngInfo = NULL;
	    actInfo = NULL;

	    state = NEWSGROUP;
	  }
	  break;

	default:
	  assert (0);
	}
      }
      while (rereadLine);
    }

    if (ngInfo && ngInfo->wildType == WILD_TYPE_NONE) {
      fprintf (stderr, "%s: %s: unexpected EOF reached\n", PROGRAM_NAME,
	       basename (conf));
      success = 0;
    }

    fclose (rcFile);
  }
  else
    fprintf (stderr, "%s: unable to open file '%s'\n", PROGRAM_NAME,
	     conf);

  return success;
}

/* Attempt to read the server configuration file. Return 1 if successful
   (or if it doesn't exist) and 0 otherwise. */
int
ReadServerConfig (const char *serverConf)
{
  int success = 1;
  FILE *servFile;
  char lineBuf[LINEBUF_LEN], *line;
  int lineNo = 0;

  int i;
  struct Server *servInfo = NULL;
  char *attr;
  int port;

  if (!(servFile = myfopen (serverConf, "r")))
    return 1; /* Just silently return if one doesn't exist. */

  while (fgets (lineBuf, sizeof (lineBuf), servFile)) {
    line = lineBuf;
    lineNo++;

    while (*line && isspace ((int) *line))
      line++;

    if (*line == '#')
      continue;

    /* Strip any trailing whitespace. */
    i = strlen (line) - 1;
    for (; i > 0 && isspace ((int) line[i]); i--);
    line[++i] = '\0';

    if (!(*line))
      continue;

    if (*line == '[') {
      /* The start of a server block. */
      line++;

      /* Find the close bracket. */
      for (i = 0; line[i] && line[i] != ']'; i++);

      if (!line[i]) {
	fprintf (stderr, "%s: %s:%d: bad server declaration\n",
		 PROGRAM_NAME, basename (serverConf), lineNo);
	success = 0;
	break;
      }

      line[i++] = '\0';

      /* Let's be anal. Make sure this is absolutely the only thing on this
	 line. */
      while (line[i] && isspace ((int) line[i]))
	i++;

      if (line[i]) {
	fprintf (stderr, "%s: %s:%d: bad server declaration\n",
		 PROGRAM_NAME, basename (serverConf), lineNo);
	success = 0;
	break;
      }

      /* Looks good. */
      if (!(servInfo = FindServer (line))) {
	/* Doesn't exist yet. Create it. */
	servInfo = AddServer (line);
      }
      assert (servInfo != NULL);
    }
    else if (!servInfo) {
      /* A server attribute without an enclosing server??? */
      fprintf (stderr,
	       "%s: %s:%d: attribute defined without server declaration\n",
	       PROGRAM_NAME, basename (serverConf), lineNo);
      success = 0;
      break;
    }
    else {
      /* Some sort of server attribute. Try to delineate it. */
      attr = line;
      while (*line && !isspace ((int) *line))
	line++;

      if (*line) {
	*line = '\0';
	line++;
	/* Eat up all spaces to find our argument. */
	while (*line && isspace ((int) *line))
	  line++;
      }

      if (!strcasecmp (attr, "port") && *line && (port = atoi (line)) > 0 &&
	  port < 65536) {
	servInfo->port = port;
      }
      else if (!strcasecmp (attr, "host") && *line) {
	if (servInfo->host)
	  free (servInfo->host);
	servInfo->host = mystrdup (line);
      }
      else if (!strcasecmp (attr, "config") && *line) {
	if (servInfo->rcFile)
	  free (servInfo->rcFile);
	servInfo->rcFile = mystrdup (line);
      }
      else if (!strcasecmp (attr, "newsrc") && *line) {
	if (servInfo->newsRCFile)
	  free (servInfo->newsRCFile);
	servInfo->newsRCFile = mystrdup (line);
      }
      else if ((!strcasecmp (attr, "user") || !strcasecmp (attr, "username"))
	       && *line) {
	if (servInfo->user)
	  free (servInfo->user);
	servInfo->user = mystrdup (line);
      }
      else if ((!strcasecmp (attr, "pass") || !strcasecmp (attr, "password"))
	       && *line) {
	if (servInfo->password)
	  free (servInfo->password);
	servInfo->password = mystrdup (line);
      }
      else if (!strcasecmp (attr, "preconnect") && *line) {
	if (servInfo->preconnectCmd)
	  free (servInfo->preconnectCmd);
	servInfo->preconnectCmd = mystrdup (line);
      }
      else if (!strcasecmp (attr, "postconnect") && *line) {
	if (servInfo->postconnectCmd)
	  free (servInfo->postconnectCmd);
	servInfo->postconnectCmd = mystrdup (line);
      }
      else {
	fprintf (stderr, "%s: %s:%d: invalid server attribute\n",
		 PROGRAM_NAME, basename (serverConf), lineNo);
	success = 0;
	break;
      }
    }
  }

  fclose (servFile);

  return success;
}
