/* -*- C -*-
 * Copyright (c) 1999, 2000 Allan Saddi <asaddi@philosophysw.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: newsfish.c,v 1.40 2001/06/19 02:46:04 asaddi Exp $
 */

#include <sys/types.h>
#ifdef HAVE_SYS_FILIO_H
#include <sys/filio.h>
#endif /* HAVE_SYS_FILIO_H */
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/stat.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <netdb.h>
#include <pwd.h>
#include <regex.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>

#include <assert.h>

#include "util.h"
#include "list.h"
#include "newsfish.h"
#include "nntppipeline.h"
#include "config.h"
#include "version.h"

#ifndef lint
static const char rcsid[] =
	"$Id: newsfish.c,v 1.40 2001/06/19 02:46:04 asaddi Exp $";
#endif /* !lint */

struct catchupPattern
{
  struct catchupPattern *next;
  char *name;
};

#define LINEBUF_LEN 1024

int BeVerbose;
unsigned long GlobalSizeLimit;
int DefaultPort;
int KeepGoing;

static FILE *nntpInFile, *nntpOutFile;

int authServer (const struct Server *servInfo);

/* Utility routines so we don't have to repeat ourself over and over again. */
void *
mymalloc (size_t size)
{
  void *mem;

  if (!(mem = malloc (size))) {
    fprintf (stderr, "%s: malloc() failed\n", PROGRAM_NAME);
    exit (1);
  }

  return mem;
}

char *
mystrdup (const char *str)
{
  char *s;

  assert (str != NULL);

  if (!(s = strdup (str))) {
    fprintf (stderr, "%s: strdup() failed\n", PROGRAM_NAME);
    exit (1);
  }

  return s;
}

FILE *
myfopen (const char *path, const char *mode)
{
  struct passwd *pwd;
  char *homeDir, newPath[PATH_MAX];
  int i;

  assert (path != NULL);
  assert (mode != NULL);

  if (path[0] == '/' ||
      !strncmp (path, "./", 2) ||
      !strncmp (path, "../", 3)) {
    /* Absolute or specific relative path. Do nothing else. */
    return fopen (path, mode);
  }

  /* Otherwise, try to get the user's home directory. */
  homeDir = NULL;
  if ((pwd = getpwuid (getuid ())))
    homeDir = pwd->pw_dir;
  if (!homeDir)
    homeDir = "/";

  i = strlcpy (newPath, homeDir, sizeof (newPath));
  if (i >= sizeof (newPath)) /* Strange... but just in case. */
    return NULL; 
  if (newPath[i] != '/') {
    /* Add a separator, if need be. */
    newPath[i++] = '/';
    if (i >= sizeof (newPath))
      return NULL;
  }
  /* Finally, tack on the path we were given. */
  if (strlcpy (&newPath[i], path, sizeof (newPath) - i) >=
      (sizeof (newPath) - 1))
    return NULL;

  return fopen (newPath, mode);
}

/* NNTP stuff. */

/* Given an NNTP server, connect to it and set up our file streams. */
static int
nntpConnect (const struct Server *servInfo, int port)
{
  const char *nntpServer;
  int nntpSock, nntpSock2;
  struct sockaddr_in nntpSockAddr;
  struct hostent *he;

  char lineBuf[LINEBUF_LEN];

  assert (servInfo != NULL);

  nntpServer = servInfo->host ? servInfo->host : servInfo->name;
  port = port ? port : servInfo->port;

  nntpSockAddr.sin_family = AF_INET;
  nntpSockAddr.sin_port = htons (port);

  /* Try to resolve nntpServer. */
  if (!inet_aton (nntpServer, &nntpSockAddr.sin_addr)) {
    /* Not a valid IP address. Try DNS. */
    if ((he = gethostbyname (nntpServer)))
      nntpSockAddr.sin_addr = *((struct in_addr *) (he->h_addr_list[0]));
    else {
      fprintf (stderr, "%s: %s: unknown host\n", PROGRAM_NAME, nntpServer);
      return 0;
    }
  }
 
  /* Create a socket, connect it to the server. */
  if ((nntpSock = socket (AF_INET, SOCK_STREAM, PF_UNSPEC)) == -1) {
    perror (PROGRAM_NAME);
    return 0;
  }

  if (BeVerbose)
    fprintf (stderr, "Connecting to %s:%d... ", nntpServer, port);

  if (connect (nntpSock, (struct sockaddr *) &nntpSockAddr,
	       sizeof (nntpSockAddr))) {
    perror (PROGRAM_NAME);
    return 0;
  }

  if (BeVerbose)
    fprintf (stderr, "ok\n");

  /* Set up the socket with stdio streams. */
  if ((nntpSock2 = dup (nntpSock)) == -1) {
    perror (PROGRAM_NAME);
    return 0;
  }

  if (!(nntpInFile = fdopen (nntpSock, "r"))) {
    perror (PROGRAM_NAME);
    return 0;
  }
#ifndef SETVBUF_REVERSED
  setvbuf (nntpInFile, NULL, _IOLBF, 0);
#else /* !SETVBUF_REVERSED */
  setvbuf (nntpInFile, _IOLBF, NULL, 0);
#endif /* !SETVBUF_REVERSED */

  if (!(nntpOutFile = fdopen (nntpSock2, "w"))) {
    perror (PROGRAM_NAME);
    return 0;
  }
#ifndef SETVBUF_REVERSED
  setvbuf (nntpOutFile, NULL, _IOLBF, 0);
#else /* !SETVBUF_REVERSED */
  setvbuf (nntpOutFile, _IOLBF, NULL, 0);
#endif /* !SETVBUF_REVERSED */

  /* Finally, check the reponse of the server. */
  fgets (lineBuf, LINEBUF_LEN, nntpInFile);
  if (!ferror (nntpInFile)) {
    if (strncmp (lineBuf, "200", 3) && strncmp (lineBuf, "201", 3)) {
      fprintf (stderr, "%s: bad response from NNTP server\n", PROGRAM_NAME);
      return 0;
    }

    if (BeVerbose)
      fprintf (stderr, "Response from NNTP server:\n  %s", lineBuf);
  }
  else {
    fprintf (stderr, "%s: error talking to NNTP server -- %s\n",
	     PROGRAM_NAME, strerror (errno));
    return 0;
  }

  /* Try to authenticate with server. */
  if (!authServer (servInfo))
    return 0;

  return 1;
}

/* Disconnect from the NNTP server. */
static void
nntpDisconnect (void)
{
  assert (nntpOutFile != NULL);
  fprintf (nntpOutFile, "QUIT\r\n");
  fclose (nntpOutFile);
  assert (nntpInFile != NULL);
  fclose (nntpInFile);

  if (BeVerbose)
    fprintf (stderr, "Disconnected from NNTP server\n");
}

/* core newsfish stuff. */

/* Get rid of any trailing carriage returns/linefeeds. */
static void
eraseCRLF (char *line)
{
  int i;

  i = strlen (line) - 1;
  while (i >= 0 && (line[i] == '\r' || line[i] == '\n'))
    line[i--] = '\0';
}

/* Authenticate ourself with the server, if needed. */
int
authServer (const struct Server *servInfo)
{
  char respBuf[LINEBUF_LEN];
  int moreAuth;

  assert (servInfo != NULL);

  if (servInfo->password) {
    if (BeVerbose)
      fprintf (stderr, "Authenticating with NNTP server...\n");

    /* This server requires authentication. */
    moreAuth = 1;
    if (servInfo->user) {
      fprintf (nntpOutFile, "AUTHINFO USER %s\r\n", servInfo->user);
      fgets (respBuf, sizeof (respBuf), nntpInFile);
      if (!ferror (nntpInFile)) {
	moreAuth = strncmp (respBuf, "281", 3);
	if (moreAuth && strncmp (respBuf, "381", 3)) {
	  eraseCRLF (respBuf);
	  fprintf (stderr, "%s: bad NNTP server response -- %s\n",
		   PROGRAM_NAME, respBuf);
	  return 0;
	}
      }
      else {
	fprintf (stderr, "%s: error talking to NNTP server -- %s\n",
		 PROGRAM_NAME, strerror (errno));
	return 0;
      }
    }

    if (moreAuth) {
      fprintf (nntpOutFile, "AUTHINFO PASS %s\r\n", servInfo->password);
      fgets (respBuf, sizeof (respBuf), nntpInFile);
      if (!ferror (nntpInFile)) {
	if (strncmp (respBuf, "281", 3)) {
	  eraseCRLF (respBuf);
	  fprintf (stderr, "%s: bad NNTP server response -- %s\n",
		   PROGRAM_NAME, respBuf);
	  return 0;
	}
      }
      else {
	fprintf (stderr, "%s: error talking to NNTP server -- %s\n",
		 PROGRAM_NAME, strerror (errno));
	return 0;
      }
    }
  }

  return 1;
}

/* Match a newsgroup name with a wildcard newsgroup structure. See note
   above about what kinds of patterns we support. */
int
MatchWildNewsgroup (const char *newsgroup, struct Newsgroup *ngInfo)
{
  int targetLen, patLen;

  assert (newsgroup != NULL);
  assert (ngInfo != NULL);
  assert (ngInfo->wildType != WILD_TYPE_NONE);

  targetLen = strlen (newsgroup);
  patLen = strlen (ngInfo->name);

  /* Easy case first. Always matches. */
  if (ngInfo->wildType == WILD_TYPE_ALL)
    return 1;

  /* Make sure the target is as long as the pattern. */
  if (targetLen < patLen)
    return 0; /* No possible way it could match. */

  switch (ngInfo->wildType) {
  case WILD_TYPE_PREFIX:
    /* Prefix match? */
    if (!strncmp (newsgroup, ngInfo->name, patLen))
      return 1;
    break;
  case WILD_TYPE_SUFFIX:
    /* Suffix match? */
    if (!strcmp (&newsgroup[targetLen - patLen], ngInfo->name))
      return 1;
    break;
  case WILD_TYPE_MIDDLE:
    /* Middle match? */
    if (strstr (newsgroup, ngInfo->name))
      return 1;
    break;
  default:
    assert (0);
  }
	
  /* No match. */
  return 0;
}

/* Our lock routines, all based on lock files. */
static int
attemptLock (const char *lockName)
{
  int fakefd;
  struct stat fakesb, locksb;
  int acquired = 0;
  char fakeName[PATH_MAX];

  assert (lockName != NULL);

  strlcpy (fakeName, lockName, sizeof (fakeName));
  strlcat (fakeName, ".XXXXXX", sizeof (fakeName));

  if ((fakefd = mkstemp (fakeName)) != -1) {
    fchmod (fakefd, 0666);
    close (fakefd);

    if (!link (fakeName, lockName)) {
      /* Link successful. Verify that we actually did succeed. */
      if (!stat (fakeName, &fakesb) && !stat (lockName, &locksb)) {
	if (fakesb.st_dev == locksb.st_dev &&
	    fakesb.st_ino == locksb.st_ino) {
	  /* Success! */
	  acquired = 1;
	}
      }
    }

    unlink (fakeName);
  }
  else
    acquired = -1; /* Something is fatally wrong. */

  return acquired;
}

static int
acquireLock (const char *name)
{
  int res;
  struct stat sb;
  time_t now;
  char lockName[PATH_MAX];

  strlcpy (lockName, name, sizeof (lockName));
  strlcat (lockName, ".lock", sizeof (lockName));

  while (!(res = attemptLock (lockName))) {
    /* If the lock is stale (5 mins), just remove it. */
    if (!stat (lockName, &sb)) {
      time (&now);
      if (now >= (sb.st_mtime + 300))
	unlink (lockName);
    }

    /* Sleep for a second before we try again. */
    sleep (1);
  }

  return res != -1;
}

static void
releaseLock (const char *name)
{
  char lockName[PATH_MAX];

  strlcpy (lockName, name, sizeof (lockName));
  strlcat (lockName, ".lock", sizeof (lockName));
  unlink (lockName);
}

static void
writeArticle (FILE *ngFile,
	      const char *header, int headerLen,
	      const char *body, int bodyLen)
{
  time_t currentTime;

  assert (ngFile != NULL);
  assert (header != NULL);
  assert (body != NULL);

  /* Write out the "From " header (for mailbox formatting). */
  time (&currentTime);
  fprintf (ngFile, "From newsfish@newsfish %s",
	   ctime (&currentTime));

  /* Write out the header. */
  fwrite (header, 1, headerLen, ngFile);

  /* Blank lines to separate header & body. */
  fprintf (ngFile, "\n\n");

  /* Write out body. */
  fwrite (body, 1, bodyLen, ngFile);

  /* More blank lines. */
  fprintf (ngFile, "\n\n");
}

/* Process a pipe action. Returns 0 for failure, and 1 for success. */
static int
processPipeAction (struct Action *actInfo,
		   const char *header, int headerLen,
		   const char *body, int bodyLen)
{
  FILE *ngFile;
  int success = 0;

  assert (actInfo != NULL);
  assert (actInfo->type == ACTION_TYPE_PIPE);
  assert (header != NULL);
  assert (body != NULL);

  /* Make sure our output buffers are empty. */
  fflush (stdout);
  fflush (stderr);

  /* Attempt to open the pipe to the program. */
  if (!(ngFile = popen (actInfo->file, "w"))) {
    fprintf (stderr, "%s: unable to open pipe to '%s'\n",
	     PROGRAM_NAME, actInfo->file);
    return 0;
  }

  writeArticle (ngFile, header, headerLen, body, bodyLen);

  /* If there were any errors, let the user know. */
  if (ferror (ngFile))
    fprintf (stderr, "%s: error writing to '%s'\n", PROGRAM_NAME,
	     actInfo->file);
  else
    success = 1;

  /* Print a warning if the pipe returned non-zero. */
  if (pclose (ngFile)) {
    fprintf (stderr, "%s: pipe '%s' terminated abnormally\n",
	     PROGRAM_NAME, actInfo->file);
    success = 0;
  }

  return success;
}

/* Process a file action. Returns 0 for failure, -1 for quit (e.g. when the
   destination file is too large, and 1 for success. */
static int
processFileAction (struct Action *actInfo,
		   const char *header, int headerLen,
		   const char *body, int bodyLen)
{
  struct stat sb;
  FILE *ngFile;

  assert (actInfo != NULL);
  assert (actInfo->type == ACTION_TYPE_FILE);
  assert (header != NULL);
  assert (body != NULL);

  /* First, try to get a lock on the file. */
  if (!acquireLock (actInfo->file)) {
    fprintf (stderr, "%s: unable to open file '%s'\n",
	     PROGRAM_NAME, actInfo->file);
    return 0;
  }

  /* Attempt to stat the file to get its size. */
  if (!stat (actInfo->file, &sb)) {
    /* See if we have any room left. */
    if (actInfo->sizeLimit && sb.st_size > actInfo->sizeLimit) {
      fprintf (stderr, "%s: file size limit reached on '%s'\n",
	       PROGRAM_NAME, actInfo->file);
      releaseLock (actInfo->file);
      return -1;
    }
  }
  else if (errno != ENOENT) {
    /* If the file doesn't exist, just treat it as zero-size.
       Otherwise... */
    fprintf (stderr, "%s: unable to open file '%s'\n",
	     PROGRAM_NAME, actInfo->file);
    releaseLock (actInfo->file);
    return 0;
  }
  /* Attempt to open the file. */
  if (!(ngFile = fopen (actInfo->file, "a"))) {
    fprintf (stderr,
	     "%s: unable to open file '%s' for appending\n",
	     PROGRAM_NAME, actInfo->file);
    releaseLock (actInfo->file);
    return 0;
  }

  writeArticle (ngFile, header, headerLen, body, bodyLen);

  /* If anything goes wrong while writing, get outta here. */
  if (ferror (ngFile)) {
    fprintf (stderr, "%s: error writing to '%s'\n",
	     PROGRAM_NAME, actInfo->file);
    fclose (ngFile);
    releaseLock (actInfo->file);
    return 0;
  }

  fclose (ngFile);
  releaseLock (actInfo->file);

  return 1;
}

/* Process an action. Returns 0 for failure, -1 for quit (e.g. when the
   destination file is too large, and 1 for success. */
int
ProcessAction (struct Action *actInfo,
	       const char *header, int headerLen,
	       const char *body, int bodyLen)
{
  assert (actInfo != NULL);
  assert (header != NULL);
  assert (body != NULL);

  if (!actInfo->file[0]) {
    /* Got no file, got nothing to do! */
    return 1;
  }

  switch (actInfo->type) {
  case ACTION_TYPE_FILE:
    return processFileAction (actInfo, header, headerLen, body, bodyLen);
  case ACTION_TYPE_PIPE:
    return processPipeAction (actInfo, header, headerLen, body, bodyLen);
  default:
    /* I would normally just assert(0), but we'll do nothing for now. */
    return 1;
  }
}

/* The absolute, real, core newsfish thing. Scans all unread articles in
   a newsgroup. Appropriate action is taken for any articles that match
   a set of conditions. Returns 0 if something went fatally wrong,
   1 if all is well (or a recoverable error occured), 2 if the user wants
   to quit. */
static int
processNewsgroup (struct Newsgroup *ngInfo, int maxReq)
{
  char lineBuf[LINEBUF_LEN];
  int articleLow, articleHigh;
  int i;

  assert (ngInfo->name[0] != '\0');

  /* If this newsgroup has no actions, just skip it. */
  if (IS_LIST_EMPTY (&ngInfo->actionList))
    return 1;

  if (BeVerbose)
    fprintf (stderr, "Processing group %s...\n", ngInfo->name);

  /* Issue a GROUP command to switch to that newsgroup. */
  fprintf (nntpOutFile, "GROUP %s\r\n", ngInfo->name);
  fgets (lineBuf, LINEBUF_LEN, nntpInFile);
  if (!ferror (nntpInFile)) {
    if (!strncmp (lineBuf, "211", 3)) {
      if (sscanf (lineBuf, "211 %*d %d %d", &articleLow, &articleHigh) != 2) {
	eraseCRLF (lineBuf);
	fprintf (stderr, "%s: bad NNTP server response -- %s\n",
		 PROGRAM_NAME, lineBuf);
	return 1;
      }
    }
    else if (!strncmp (lineBuf, "411", 3)) {
      fprintf (stderr, "%s: %s: newsgroup does not exist\n", PROGRAM_NAME,
	       ngInfo->name);
      return 1;
    }
    else {
      eraseCRLF (lineBuf);
      fprintf (stderr, "%s: bad NNTP server response -- %s\n",
	       PROGRAM_NAME, lineBuf);
      return 1;
    }
  }
  else {
    fprintf (stderr, "%s: error talking to NNTP server -- %s\n",
	     PROGRAM_NAME, strerror (errno));
    return 0;
  }

  /* Adjust our article pointer. */
  if (ngInfo->lastArticle < articleLow)
    ngInfo->lastArticle = articleLow;
  if (ngInfo->lastArticle > articleHigh)
    ngInfo->lastArticle = articleHigh + 1;

  if (BeVerbose)
    fprintf (stderr, "Reading articles from range %d-%d, starting at %d...\n",
	     articleLow, articleHigh, ngInfo->lastArticle);

  /* Just in case. */
  fflush (nntpOutFile);

  /* Switch to non-blocking. */
  i = 1;
  ioctl (fileno (nntpInFile), FIONBIO, &i);
  ioctl (fileno (nntpOutFile), FIONBIO, &i);

  InitPipeline (fileno (nntpInFile), fileno (nntpOutFile), maxReq);
  PipelineEngine (ngInfo, articleHigh);
  FinishPipeline ();

  /* Return to blocking. */
  i = 0;
  ioctl (fileno (nntpInFile), FIONBIO, &i);
  ioctl (fileno (nntpOutFile), FIONBIO, &i);

  if (BeVerbose) {
    fprintf (stderr, "Done processing group %s\n", ngInfo->name);
    fprintf (stderr, "Next article to read is %d\n", ngInfo->lastArticle);
  }

  if (KeepGoing)
    return 1;
  else
    return 2;
}

static int
listNewsgroups (void)
{
  char lineBuf[LINEBUF_LEN];

  if (BeVerbose)
    fprintf (stderr, "Listing all newsgroups...\n");

  /* Issue a LIST command. */
  fprintf (nntpOutFile, "LIST\r\n");
  fgets (lineBuf, sizeof (lineBuf), nntpInFile);
  if (ferror (nntpInFile)) {
    fprintf (stderr, "%s: error talking to NNTP server -- %s\n",
	     PROGRAM_NAME, strerror (errno));
    return 0;
  }

  if (strncmp (lineBuf, "215", 3)) {
    eraseCRLF (lineBuf);
    fprintf (stderr, "%s: bad NNTP server response -- %s\n",
	     PROGRAM_NAME, lineBuf);
    return 0;
  }

  for (;;) {
    fgets (lineBuf, sizeof (lineBuf), nntpInFile);
    if (ferror (nntpInFile)) {
      fprintf (stderr, "%s: error talking to NNTP server -- %s\n",
	       PROGRAM_NAME, strerror (errno));
      return 0;
    }

    if (!strcmp (lineBuf, ".\r\n"))
      break;

    /* Just print it out verbatim. */
    eraseCRLF (lineBuf);
    printf ("%s\n", lineBuf);
  }

  return 1;
}

static int
catchupNewsgroup (struct Newsgroup *ngInfo)
{
  char lineBuf[LINEBUF_LEN];
  int articleLow, articleHigh;

  if (BeVerbose)
    fprintf (stderr, "Marking all articles in group %s...\n", ngInfo->name);

  /* Issue a GROUP command to switch to that newsgroup. */
  fprintf (nntpOutFile, "GROUP %s\r\n", ngInfo->name);
  fgets (lineBuf, LINEBUF_LEN, nntpInFile);
  if (!ferror (nntpInFile)) {
    if (!strncmp (lineBuf, "211", 3)) {
      if (sscanf (lineBuf, "211 %*d %d %d", &articleLow, &articleHigh) != 2) {
	eraseCRLF (lineBuf);
	fprintf (stderr, "%s: bad NNTP server response -- %s\n",
		 PROGRAM_NAME, lineBuf);
	return 1;
      }
    }
    else {
      eraseCRLF (lineBuf);
      fprintf (stderr, "%s: bad NNTP server response -- %s\n",
	       PROGRAM_NAME, lineBuf);
      return 1;
    }
  }
  else {
    fprintf (stderr, "%s: error talking to NNTP server -- %s\n",
	     PROGRAM_NAME, strerror (errno));
    return 0;
  }

  /* Adjust our article pointer. */
  ngInfo->lastArticle = articleHigh + 1;

  return 1;
}

/* If catchList is NULL, will catch up on all defined newsgroups. */
static int
catchup (struct catchupPattern *catchList)
{
  struct Newsgroup *ngInfo;

  if (catchList) {
    do {
      if ((ngInfo = FindNewsgroup (catchList->name))) {
	if (!catchupNewsgroup (ngInfo))
	  return 0;
      }
      else
	fprintf (stderr, "%s: group %s not defined in filters; skipping...\n",
		 PROGRAM_NAME, catchList->name);
      catchList = catchList->next;
    }
    while (catchList);
  }
  else {
    FOR_EACH_NODE_IN_LIST (&NewsgroupList, ngInfo)
      if (!IS_LIST_EMPTY (&ngInfo->actionList))
	if (!catchupNewsgroup (ngInfo))
	  return 0;
  }

  return 1;
}

/* Everything else. */

static void
printUsage (void)
{
  fprintf (stderr,
	   "Usage: %s [-CFVhlv] [-f <config>] [[-c <group>] ...] [-m <size>]\n                [-n <maxReq>] [-p <port>] [<NNTP server>]\n",
	   PROGRAM_NAME);
}

static void
printHelp (void)
{
  fprintf (stderr, "\n\
Command-line summary:\n\
  -C\t\tMark all articles in all newsgroups as read.\n\
  -F\t\tConnect even if preconnect command fails.\n\
  -V\t\tDisplay version information.\n\
  -h\t\tDisplay this summary.\n\
  -l\t\tList all available newsgroups.\n\
  -v\t\tBe verbose.\n\
  -f <config>\tRead an alternate server configuration file.\n\
  \t\t(default: $HOME/.newsfish-servers)\n\
  -c <group>\tMark all articles in <group> as read.\n\
  -m <size>\tLimit files to <size> bytes.\n\
  -n <maxReq>\tSet maximum allowable outstanding NNTP requests.\n\
  \t\t(default: 1)\n\
  -p <port>\tConnect to a different port. (default: 119)\n\
  <NNTP server>\tThe name of the NNTP server. This overrides\n\
\t\tthe NNTPSERVER environment variable.\n");
}

RETSIGTYPE
die (int foo)
{
  fprintf (stderr, "%s: signal caught... exiting\n", PROGRAM_NAME);
  KeepGoing = 0;
}

int
main (int argc, char *argv[])
{
  int failure = 0;
  char *nntpServer;
  struct Newsgroup *ngInfo;
  int procNGresult;

  int c;
  char *serverConf;
  char *mod;
  int portSwitch;
  int ignorePreconnectFail;
  int catchAll;
  struct catchupPattern *catchList, *cp;
  int doList;
  int maxReq;

  struct servent *se;

  struct Server *servInfo;
  char *conf, *newsrc;

  BeVerbose = 0;
  GlobalSizeLimit = 0L; /* No limit. */
  serverConf = NULL;
  portSwitch = 0;
  ignorePreconnectFail = 0;
  catchAll = 0;
  catchList = NULL;
  doList = 0;
  maxReq = 1;

  while ((c = getopt (argc, argv, "CFVc:f:hlm:n:p:v")) != -1) {
    switch (c) {
    case 'C':
      catchAll++;
      break;
    case 'F':
      ignorePreconnectFail++;
      break;
    case 'V':
      fprintf (stderr, "%s version " VERSION_STRING " (" VERSION_DATE ")\n",
	       PROGRAM_NAME);
      exit (1);
      break;
    case 'c':
      cp = mymalloc (sizeof (*cp));
      cp->next = catchList;
      cp->name = optarg; /* should be safe */
      catchList = cp;
      break;
    case 'f':
      serverConf = optarg;
      break;
    case 'h':
      printUsage ();
      printHelp ();
      exit (1);
      break;
    case 'l':
      doList++;
      break;
    case 'm':
      GlobalSizeLimit = strtoul (optarg, &mod, 0);
      if (mod == optarg) {
	fprintf (stderr, "%s: illegal numeric value for -m\n",
		 PROGRAM_NAME);
	exit (1);
      }
      switch (*mod) {
      case 'b':
	GlobalSizeLimit *= 512;
	mod++;
	break;
      case 'k':
	GlobalSizeLimit *= 1024;
	mod++;
	break;
      case 'm':
	GlobalSizeLimit *= 1048576;
	mod++;
	break;
      }
      if (*mod) {
	fprintf (stderr, "%s: illegal numeric value for -m\n",
		 PROGRAM_NAME);
	exit (1);
      }
      break;
    case 'n':
      maxReq = atoi (optarg);
      if (maxReq < 1) {
	fprintf (stderr, "%s: bad argument for -n\n",
		 PROGRAM_NAME);
      }
      break;
    case 'p':
      portSwitch = atoi (optarg);
      if (portSwitch < 1 || portSwitch > 65535) {
	fprintf (stderr, "%s: bad argument for -p\n",
		 PROGRAM_NAME);
	exit (1);
      }
      break;
    case 'v':
      BeVerbose = 1;
      break;
    case '?':
    default:
      printUsage ();
      exit (1);
    }
  }
  argc -= optind;
  argv += optind;

  if (argc > 1) {
    printUsage ();
    exit (1);
  }

  /* Figure out which server we connect to. */
  if (argc == 1)
    nntpServer = argv[0];
  else {
    /* Figure out the NNTP server. */
    nntpServer = getenv ("NNTPSERVER");
    if (!nntpServer || !(*nntpServer)) {
      fprintf (stderr, "%s: environment variable NNTPSERVER undefined\n",
	       PROGRAM_NAME);
      exit (1);
    }
  }

  if (BeVerbose && GlobalSizeLimit)
    fprintf (stderr, "Global file size limit set to %lu bytes\n",
	     GlobalSizeLimit);

  InitList (&ServerList);
  InitList (&NewsgroupList);
  InitList (&WildNewsgroupList);

  /* Attempt to resolve the default NNTP port. */
  if ((se = getservbyname ("nntp", "tcp")))
    DefaultPort = ntohs (se->s_port);
  else {
    /* We used to complain and exit. We'll just assume it's 119. */
    DefaultPort = 119;
  }

  if (!serverConf)
    serverConf = SERVERCONF_FILE_NAME;

  if (BeVerbose)
    fprintf (stderr, "Server config file is '%s'\n", serverConf);

  /* Attempt to read our server config file, if any. */
  if (!ReadServerConfig (serverConf))
    exit (1); /* Something went wrong. */

  /* Try to find this server's attributes. If not found, just create
     a blank template. */
  if (!(servInfo = FindServer (nntpServer))) {
    servInfo = AddServer (nntpServer);
  }
  assert (servInfo != NULL);

  conf = servInfo->rcFile ? servInfo->rcFile : RC_FILE_NAME;

  if (BeVerbose)
    fprintf (stderr, "Newsfish config file is '%s'\n", conf);

  /* Try to read the config file for this server. */
  if (!ReadConfig (conf))
    exit (1); /* Something went wrong. */

  newsrc = servInfo->newsRCFile ? servInfo->newsRCFile : NEWSRC_FILE_NAME;

  if (BeVerbose)
    fprintf (stderr, "Newsrc file is '%s'\n", newsrc);

  /* We don't care if the following is successful or not. */
  ReadNewsRC (newsrc);

  /* Run the preconnect command, if any. */
  if (servInfo->preconnectCmd) {
    if (BeVerbose)
      fprintf (stderr, "Executing preconnect command...\n");

    if (system (servInfo->preconnectCmd)) {
      fprintf (stderr,
	       "%s: preconnect command terminated abnormally\n",
	       PROGRAM_NAME);

      if (!ignorePreconnectFail)
	exit (1);
    }
  }

  KeepGoing = 1;

  /* Catch a few signals so we can quit cleanly. */
  signal (SIGHUP, &die);
  signal (SIGINT, &die);
  signal (SIGQUIT, &die);

  /* Connect to the NNTP server. */
  if (nntpConnect (servInfo, portSwitch)) {
    /* If the user wants to catch up any groups, do that instead. */
    if (catchAll || catchList)
      catchup (catchAll ? NULL : catchList);
    else if (doList) {
      /* If the user wants a newsgroup listing, do that instead. */
      failure = !listNewsgroups ();
    }
    else {
      /* Loop through each newsgroup and process it. */
      FOR_EACH_NODE_IN_LIST (&NewsgroupList, ngInfo) {
	procNGresult = processNewsgroup (ngInfo, maxReq);
	if (!procNGresult) {
	  failure = 1;
	  break;
	}
	else if (procNGresult == 2) /* Time to die. */
	  break;
      }
    }

    nntpDisconnect ();

    /* Save the latest last article pointers. */
    if (BeVerbose)
      fprintf (stderr, "Saving newsrc to '%s'\n", newsrc);

    WriteNewsRC (newsrc);
  }

  /* Run the postconnect command, if any. */
  if (servInfo->postconnectCmd) {
    if (BeVerbose)
      fprintf (stderr, "Executing postconnect command...\n");

    if (system (servInfo->postconnectCmd))
      fprintf (stderr, "%s: postconnect command terminated abnormally\n",
	       PROGRAM_NAME);
  }

  exit (failure);
}
