/* -*- C -*-
 * Copyright (c) 2000 Allan Saddi <asaddi@philosophysw.com>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	$Id: nntppipeline.c,v 1.27 2000/08/03 04:24:49 asaddi Exp $
 */

#include <sys/types.h>
#include <sys/time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>

#include "list.h"
#include "newsfish.h"
#include "config.h"

#ifndef lint
static const char rcsid[] =
	"$Id: nntppipeline.c,v 1.27 2000/08/03 04:24:49 asaddi Exp $";
#endif /* !lint */

#define NNTP_BUFFER_SIZE 4096
#define HEADER_BUFFER_SIZE 1024
#define BODY_BUFFER_SIZE 65536
#define HEADER_BUCKET_SIZE 251

struct matchedHeader {
  struct matchedHeader *next;
  int articleNum;
  int headerLen;
  const char *header;
  int matchCount;
  struct Action **actInfo;
};

struct bodyRequest {
  struct Node node;
  int articleNum;
};

struct wildNewsgroup {
  struct Node node;
  struct Newsgroup *ngInfo;
};

static int inSock, outSock;
static int inPos, outPutPos, outGetPos;
static char inBuf[NNTP_BUFFER_SIZE];
static char outBuf[NNTP_BUFFER_SIZE];

static int maxOutstanding, ackReqCount;

static int headerBufPos, bodyBufPos;
static int headerBufLen = HEADER_BUFFER_SIZE;
static int bodyBufLen = BODY_BUFFER_SIZE;
static char *headerBuf = NULL, *bodyBuf = NULL;

static enum {
  STATE_LOOKING,
  STATE_HEADER,
  STATE_BODY,
} processState;

static int articleProcess;

static struct matchedHeader *matchedHeaders[HEADER_BUCKET_SIZE];

static struct List bodyReqQueue, matchedWildList;

static int lastExaminedArticle, lastRequestedArticle, lastSavedArticle;

static void
processBody (int articleNum, const char *body, int bodyLen)
{
  unsigned long hash;
  struct matchedHeader *mh, *nmh, *last;
  int i;
  int res;

  hash = articleNum % HEADER_BUCKET_SIZE;
  mh = matchedHeaders[hash];
  while (mh) {
    if (mh->articleNum == articleNum)
      break;
    mh = mh->next;
  }

  /* If we didn't find it, something's wrong. */
  assert (mh != NULL);

  for (i = 0; i < mh->matchCount; i++) {
    res = ProcessAction (mh->actInfo[i], mh->header, mh->headerLen,
			 body, bodyLen);
    if (res < 1)
      break;
  }


  if (res < 1)
    KeepGoing = 0;
  else {
    lastSavedArticle = articleNum;

    if (BeVerbose)
      fprintf (stderr, "done!\n");
  }

  if ((nmh = matchedHeaders[hash]) == mh)
    matchedHeaders[hash] = mh->next;
  else {
    /* Since the server most likely processes our requests in order,
       we should never have to remove a matchedHeader out-of-order.
       Though, just in case... */
    last = nmh;
    nmh = nmh->next;
    while (nmh && nmh != mh) {
      last = nmh;
      nmh = nmh->next;
    }
    if (nmh)
      last->next = mh->next;
  }

  free (mh->actInfo);
  free ((char *) mh->header);
  free (mh);
}

static int
processHeader (struct Newsgroup *ngInfo, int articleNum,
	       const char *header, int headerLen)
{
  int wildMatch;
  struct wildNewsgroup *wng;

  int match;
  struct Action *actInfo;
  struct Condition *condInfo;
  int reMatch;

  struct matchedHeader *mh;
  int i;
  unsigned long hash;

  struct bodyRequest *br;

  /* Assume all wildcard newsgroups match. */
  wildMatch = 1;

  /* See if any wildcard newsgroups match the current newsgroup first. */
  FOR_EACH_NODE_IN_LIST (&matchedWildList, wng) {
    /* There should only be one action for this wildcard newsgroup.
       Try to match all conditions. */
    actInfo = (struct Action *) wng->ngInfo->actionList.head;
    assert (actInfo != NULL);

    /* Compare with each condition. */
    FOR_EACH_NODE_IN_LIST (&actInfo->conditionList, condInfo) {
      reMatch = !regexec (&condInfo->regexp, header, 0, NULL, 0);
      if (!reMatch && !condInfo->invert)
	break; /* Didn't match and we weren't inverting. */
      else if (reMatch && condInfo->invert)
	break; /* Matched and we were inverting. */
    }

    /* If we broke out of the loop, no match. */
    if (condInfo->node.next) {
      wildMatch = 0;
      break;
    }
  }

  if (!wildMatch) {
    /* No match. No need to check newsgroup-specific actions. */
    if (BeVerbose)
      fprintf (stderr, "no match. (general)\n");

    lastExaminedArticle = articleNum;

    return 0;
  }

  /* So we know if at least one action matched. */
  match = 0;

  /* Loop through each action for this newsgroup. */
  FOR_EACH_NODE_IN_LIST (&ngInfo->actionList, actInfo) {
    actInfo->match = 1;

    /* Compare with each condition. */
    FOR_EACH_NODE_IN_LIST (&actInfo->conditionList, condInfo) {
      reMatch = !regexec (&condInfo->regexp, header, 0, NULL, 0);
      if (!reMatch && !condInfo->invert)
	break; /* Didn't match and we weren't inverting. */
      else if (reMatch && condInfo->invert)
	break; /* Matched and we were inverting. */
    }

    /* If we broke out of the loop, this filter set couldn't have
       matched. */
    if (condInfo->node.next)
      actInfo->match = 0;
    else
      match++; /* We matched at least one. */
  }

  if (match) {
    /* Got a match. Remember all matching Actions and save the header. */
    mh = mymalloc (sizeof (*mh));

    mh->articleNum = articleNum;
    mh->headerLen = headerLen;
    mh->header = header;
    mh->matchCount = match;

    mh->actInfo = mymalloc (sizeof (*mh->actInfo) * match);

    i = 0;
    FOR_EACH_NODE_IN_LIST (&ngInfo->actionList, actInfo) {
      if (actInfo->match) {
	assert (i < match);
	mh->actInfo[i++] = actInfo;
      }
    }

    hash = articleNum % HEADER_BUCKET_SIZE;
    mh->next = matchedHeaders[hash];
    matchedHeaders[hash] = mh;

    /* Queue up a request to grab the body. */
    br = mymalloc (sizeof (*br));
    br->articleNum = articleNum;
    AddTail (&bodyReqQueue, &br->node);

    if (BeVerbose)
      fprintf (stderr, "match!\n");

    lastRequestedArticle = articleNum;
  }
  else if (BeVerbose)
    fprintf (stderr, "no match.\n");

  lastExaminedArticle = articleNum;

  return match != 0;
}

static void
processLine (struct Newsgroup *ngInfo, const char *line)
{
  const char *linep;
  int lineLen;
  int fromEscape, dotEscape;

  switch (processState) {
  case STATE_LOOKING:
    if (!strncmp (line, "221", 3)) {
      /* A header response. */
      linep = &line[3];
      while (*linep && isspace ((int) *linep))
	linep++;
      if (*linep) {
	articleProcess = (int) strtol (linep, NULL, 0);
	processState = STATE_HEADER;

	if (BeVerbose)
	  fprintf (stderr, "Examining article %d... ", articleProcess);
      }
    }
    else if (!strncmp (line, "222", 3)) {
      /* A body response. */
      linep = &line[3];
      while (*linep && isspace ((int) *linep))
	linep++;
      if (*linep) {
	articleProcess = (int) strtol (linep, NULL, 0);
	processState = STATE_BODY;

	if (BeVerbose)
	  fprintf (stderr, "  Saving article %d... ", articleProcess);
      }
    }
    else
      ackReqCount++;
    break;
  case STATE_HEADER:
    if (!strcmp (line, ".")) {
      /* Don't want a trailing newline. */
      if (headerBufPos)
	headerBufPos--;
      /* Terminate this buffer. */
      headerBuf[headerBufPos] = '\0';

      /* Do whatever we need to do. */
      if (processHeader (ngInfo, articleProcess, headerBuf, headerBufPos)) {
	/* Need to allocate a new headerBuf. */
	headerBufLen = HEADER_BUFFER_SIZE;
	headerBuf = mymalloc (headerBufLen);
      }

      processState = STATE_LOOKING;
      headerBufPos = 0;
      ackReqCount++;
    }
    else {
      lineLen = strlen (line);

      /* See if we have enough space in our current buffer. */
      while ((headerBufPos + lineLen + 2) > headerBufLen) {
	headerBufLen *= 2;
	if (!(headerBuf = realloc (headerBuf, headerBufLen))) {
	  fprintf (stderr, "%s: realloc() failed\n", PROGRAM_NAME);
	  exit (1);
	}
      }

      /* Copy the header line into our buffer. */
      memcpy (&headerBuf[headerBufPos], line, lineLen);
      headerBufPos += lineLen;
      /* Add a newline. */
      headerBuf[headerBufPos++] = '\n';
    }
    break;
  case STATE_BODY:
    if (!strcmp (line, ".")) {
      /* Don't want a trailing newline. */
      if (bodyBufPos)
	bodyBufPos--;
      /* Terminate this buffer. */
      bodyBuf[bodyBufPos] = '\0';

      /* Do whatever we need to do. */
      processBody (articleProcess, bodyBuf, bodyBufPos);

      processState = STATE_LOOKING;
      bodyBufPos = 0;
      ackReqCount++;
    }
    else {
      lineLen = strlen (line);

      /* Check for any special escape circumstances. */
      if ((fromEscape = !strncmp (line, "From ", 5)))
	lineLen++;
      if ((dotEscape = line[0] == '.'))
	lineLen--;

      /* See if we have enough space. */
      while ((bodyBufPos + lineLen + 2) > bodyBufLen) {
	bodyBufLen *= 2;
	if (!(bodyBuf = realloc (bodyBuf, bodyBufLen))) {
	  fprintf (stderr, "%s: realloc() failed\n", PROGRAM_NAME);
	  exit (1);
	}
      }

      /* Escape any lines beginning with "From " */
      if (fromEscape) {
	bodyBuf[bodyBufPos++] = '>'; 
	lineLen--;
      }

      /* Copy the body line into our buffer. */
      memcpy (&bodyBuf[bodyBufPos], dotEscape ? &line[1] : line,
	      lineLen);
      bodyBufPos += lineLen;
      /* Add a newline. */
      bodyBuf[bodyBufPos++] = '\n';
    }
    break;
  default:
    assert (0);
  }
}

static int
sendLine (const char *line)
{
  int lineLen, len;

  assert (line != NULL);

  /* We should not be trying to send data if we've started buffering. */
  assert (outPutPos == 0);

  lineLen = strlen (line);
  assert (lineLen > 0);

  /* Immediately attempt to send the data. */
  len = send (outSock, line, lineLen, 0);
  if (len < 0)
    len = 0; /* Assume EAGAIN or ENOBUFS */
  if (len < lineLen) {
    /* Couldn't send this request. Buffer and try again later. */
    outPutPos = lineLen - len;
    assert (outPutPos < sizeof (outBuf));
    memcpy (outBuf, &line[len], outPutPos);

    return 0;
  }

  return 1;
}

/* Do not send anymore data if we have stuff in our outgoing buffer. */
static int
readyToSend (void)
{
  return !outPutPos;
}

void
PipelineEngine (struct Newsgroup *ngInfo, int articleHigh)
{
  fd_set readSet, writeSet;
  int fdCount;
  struct timeval timeout;
  int useTimeout;
  int len, pos;
  char *start, *end, *search;

  struct bodyRequest *br;

  struct Newsgroup *wngInfo;
  struct wildNewsgroup *wng;

  char buf[256];

  FD_ZERO (&readSet);
  FD_ZERO (&writeSet);

  fdCount = inSock;
  if (outSock > fdCount)
    fdCount = outSock;
  fdCount++;

  timeout.tv_sec = 0L;
  timeout.tv_usec = 0L;

  lastExaminedArticle = lastSavedArticle = ngInfo->lastArticle - 1;
  lastRequestedArticle = 0;

  /* Find out which wildcard newsgroup match this one. */
  FOR_EACH_NODE_IN_LIST (&WildNewsgroupList, wngInfo) {
    if (MatchWildNewsgroup (ngInfo->name, wngInfo)) {
      wng = mymalloc (sizeof (*wng));
      wng->ngInfo = wngInfo;
      AddTail (&matchedWildList, &wng->node);
    }
  }

  for (;;) {
    if (KeepGoing) {
      while (ackReqCount && readyToSend () &&
	     (br = (struct bodyRequest *) RemHead (&bodyReqQueue))) {
	sprintf (buf, "BODY %d\r\n", br->articleNum);
	sendLine (buf);
	ackReqCount--;

	free (br);
      }

      if (ackReqCount && readyToSend () &&
	  ngInfo->lastArticle <= articleHigh) {
	sprintf (buf, "HEAD %d\r\n", ngInfo->lastArticle);
	sendLine (buf);
	ngInfo->lastArticle++;
	ackReqCount--;
      }
    }

    if ((!KeepGoing || ngInfo->lastArticle > articleHigh) &&
	ackReqCount == maxOutstanding) {
      /* All done here. */
      if (!KeepGoing) {
	/* Breaking out because user wants us to quit. */
	if (lastRequestedArticle == lastSavedArticle)
	  lastRequestedArticle = 0;
	if (lastRequestedArticle)
	  ngInfo->lastArticle = lastSavedArticle + 1;
	else
	  ngInfo->lastArticle = lastExaminedArticle + 1;
      }
      break;
    }

    FD_SET (inSock, &readSet);
    if (outPutPos)
      FD_SET (outSock, &writeSet);
    else
      FD_CLR (outSock, &writeSet);

    if (KeepGoing)
      useTimeout = ackReqCount && readyToSend () &&
	(!IS_LIST_EMPTY (&bodyReqQueue) || ngInfo->lastArticle <= articleHigh);
    else
      useTimeout = 0;

    if (select (fdCount, &readSet, &writeSet, NULL,
		useTimeout ? &timeout : NULL) > 0) {
      if (FD_ISSET (outSock, &writeSet)) {
	if ((len = send (outSock, &outBuf[outGetPos],
			 outPutPos - outGetPos, 0)) > 0) {
	  outGetPos += len;
	  if (outGetPos == outPutPos) {
	    /* Transmitted everything. Reset. */
	    outPutPos = outGetPos = 0;
	  }
	}
      }

      if (FD_ISSET (inSock, &readSet)) {
	if ((len = recv (inSock, &inBuf[inPos],
			 sizeof (inBuf) - inPos, 0)) > 0) {
	  inPos += len;

	  do {
	    start = inBuf;
	    pos = inPos;
	    do {
	      /* Search for a newline. */
	      search = &start[pos - len];
	      if ((end = memchr (search, '\n', len))) {
		/* Null terminate the line. */
		*(end++) = '\0';

		pos = end - start;
		/* If there was a carriage return, eliminate it. */
		if (pos > 1 && start[pos - 2] == '\r')
		  start[pos - 2] = '\0';

		/* Process the line. */
		processLine (ngInfo, start);

		/* Adjust start accordingly. */
		len -= (end - search);
		assert (len >= 0);
		start = end;
		pos = len;
	      }
	      else
		break;
	    }
	    while (pos);

	    /* Copy any remaining bytes to the beginning of the buffer. */
	    if (start != inBuf && pos > 0)
	      memcpy (inBuf, start, pos);

	    inPos = pos;

	    /* If we're about to overflow, make sure the buffer gets
	       processed. Note that the previous value of len (whatever it
	       may be) is good enough. */
	    inBuf[sizeof (inBuf) - 1] = '\n';
	  }
	  while (inPos == sizeof (inBuf));
	}

	/* FIXME: Check for len == 0 i.e. server closed connection */
      }
    }
  }
}

void
FinishPipeline (void)
{
  struct bodyRequest *br;
  struct wildNewsgroup *wngInfo;

  while ((br = (struct bodyRequest *) RemHead (&bodyReqQueue)))
    free (br);

  while ((wngInfo = (struct wildNewsgroup *) RemHead (&matchedWildList)))
    free (wngInfo);
}

void
InitPipeline (int inSock0, int outSock0, int maxOutstanding0)
{
  int i;

  inSock = inSock0;
  outSock = outSock0;

  inPos = outPutPos = outGetPos = 0;

  maxOutstanding = ackReqCount = maxOutstanding0;

  if (!headerBuf) {
    headerBuf = mymalloc (headerBufLen);
    bodyBuf = mymalloc (bodyBufLen);
  }
  headerBufPos = bodyBufPos = 0;

  processState = STATE_LOOKING;

  for (i = 0; i < HEADER_BUCKET_SIZE; i++)
    matchedHeaders[i] = NULL;

  InitList (&bodyReqQueue);
  InitList (&matchedWildList);
}
